
# Engineering principles {#engineeringprinciples}

Designing the propulsion with this tool includes different engineering disciplines. Here a brief explanation (more information in their respective sections):

- [Engine designer](#enginedesigner): calculates the performance of one individual engine based on the required thrust.
- [Propulsor integrator](#propulsionintegrator): places the engine acc. to the user's settings.
- [Nacelle designer](#nacelledesigner): calculates the nacelle geometry.
- [Pylon designer](#pylondesigner): calculates the pylon geometry.
- [Mass analyzer](#massanalyzer): calculates the mass properties (center of gravity, mass, and inertia) of engine, nacelle, and pylon.

For these five disciplines, you can choose different **methods** (or fidelities) of calculating their output. Here is an overview of the current implemented methods (details see sections):
| Discipline          | Methods                                                           |
|---------------------|-------------------------------------------------------------------|
|Engine designer      | *Rubber* (*Empirical* and *PropulsionSystem* are in preparation)  |
|Propulsor integrator | *Default*                                                         |
|Nacelle designer     | *Default*                                                         |
|Pylon designer       | *Default*                                                         |
|Mass analyzer        | *Default*                                                         |

If you want to learn more about how to configure methods or generally the settings and outputs, go to the [getting started](getting_started.md).

@important These disciplines are executed sequentially for EACH engine. That means that the engines are not aware of each other within the designing and analyzing. More information, see the [software architecture](software_architecture.md) section.


## Engine designer {#enginedesigner}

### General principles {#generalprinciples}

The **engine designer** bases its principle on the common modelling practice using 
- an _engine dataset_ (operating point **in**dependent)
- an _engine deck_ (operating point dependent)
- a _scale factor_

The _dataset_ (also called _EngineXML_) includes parameter which are independent of the flight condition such as outer engine dimensions.

The three-dimensional _engine deck_ contain engine performance data for different values of altitude $h$, Mach number $Ma$ and low-pressure engine spool speed $N1$. The most important performance parameter are thrust and fuel/energy flow. In UNICADO, the deck is split into multiple CSV files. The figure shows an example with values for thrust in kilo newton. The first block contains data for $N1=1$ for $Ma=0...0.9$ and $h=0...14000$. The second block below is for $N1=0.95$.
![](figures/deck_example_thrust.svg)

!!! note 
    Detailed information on required dataset and deck data will be updated in near future. 

The _scale factor_ is necessary because (as conceptual aircraft designer), we use the concept of a so-called _rubber engine_. That means that (depending on the method, see later) we create or assume an engine deck and provide one _scale factor_ to obtain all engine data acc. to the required thrust the engine shall provide. The figure visualized the concept:
![](figures/scale_factor.svg)

@attention &rarr; **As mentioned and highlighted in the figure**, there is ONE _scale factor_ **BUT** multiple exponents which differ depending on which property you want to use. E.g. for the diameter, the exponent is $0.5$ and for the mass its $0.4$. **So important to remember** that whenever you want to use engine data, you need to access it via the `engine` library. In the following, a brief explanation of the scaling concept will be given - however details are given in the library documentation.

So, the scaling is based on continuity principle assuming that the operating condition is constant (commonly known station numbering; assuming no pressure drop).

$ T = \dot m \cdot (V_9 - V_0) $

Therefore, thrust $T$ is proportional to the mass flow $\dot m$, which is related to the cross-sectional area $A$ of the engine. 

$ \dot m = \rho \cdot V \cdot A = \rho \cdot V \cdot \pi \frac{d}{2}^2 $

Because area $A$ is proportional to the square of the diameter $d$ , it follows that the diameter should be proportional to the square root of the scale factor. 

$ d_{new} = d_{ref} \cdot ( \frac{T_{new}}{T_{ref}} )^{0.5} $

An exemplary simplified calculation (data from the V2527-A5): the current engine provides $127.27~kN$ as sea level static thrust, but for the design only $100~kN$ are needed. The scaling factor would be $0.7857$. Assuming an initial diameter $2~m$, the new diameter would be $1.773~m$ with the scaling factor of $(0.7857)^{0.5} = 0.8864$. 

So, again, always access the engine data via the `engine` library to ensure that you have the correctly scaled data 🙂

!!! note
    Actually, the sea level static thrust is not at $N1=1$ if you compare the dataset for this engine (for 110.31kN around $N1=0.95$). So the scaling factor will be slightly lower.

### Methods description
The **engine designer** includes different methods which create/use this deck in various ways.

- *empirical*: the initial deck is calculated based on emipirical equations
- *rubber*: (most common approach) based on an existing deck (usually created with GasTurb), the deck is "rubberized"
- *propulsionsystem*: with the help of the library `propulsionsystem`, different architecture can be defined and a deck created (for more information see documentation of the library)

!!! note
    *empirical* and *propulsionsystem* is in preparation - not implemented yet!

For all these methods, the approach of using the _scale factor_ is the same (see explaination [here](#generalprinciples)). A deck is either first created or assumed and then data is drawn with the `engine` library with the scaling approach. 

## Propulsion integrator {#propulsionintegrator}
Additionally to calculating the engine performance parameter, the engine has to be placed on the aircraft. The **propulsion integrator** uses the user settings from the aircraft exchange file - the following needs to be defined:
- parent component: wing, fuselage, empennage
- x-position (aircraft coordinate system): front or rear
- y position (aircraft coordinate system): left or right
- z position (aircraft coordinate system): over, mid, under, in

### Methods description

Here, currently only one method is implemented:
 - *default* is based on a thesis of RWTH Aachen \cite{Ata10}

This method includes multiple empirical functions for different propulsion integration. These are the options that are currently implemented:
| Parent    | Lateral | Longitudinal | Vertical |
|-----------|---------|--------------|----------|
| Wing      | Right   | Front        | Under    |
| Wing      | Left    | Front        | Under    |
| Wing      | Right   | Front        | Over     |
| Wing      | Left    | Front        | Over     |
| Fuselage  | Right   | Rear         | Mid      |
| Fuselage  | Left    | Rear         | Mid      |
| Empennage | Mid     | Front        | In       |

For detailed information, it is referred to the thesis.

!!! note the implementation include currently Turbofan Kerosene only

## Nacelle designer {#nacelledesigner}
After the integration, the nacelle geometry is defined (however its actually independent of the position, so the order could be changed). 

### Methods description 

For the **nacelle designer**, only one method is implemented:

 - *default* uses the `aircraftGeometry2` library 
 
The library uses the `.dat` file defined in the _configXML_ to extrude a polygon in different sections. These sections including the origin, width, height and its profile are saved in the _acXML_. With that, every other tool can "rebuild" the geometry using the same library.

In the current implemented method, there is no differentiation between short and long ducted nacelle. It is a polygon with 3 segments (1. and 3. segments is 25% of engine length). The diameter for the 1. and 3. segment is chosen as the maximum between fan diameter, engine width or height. The 2. segments is 25% larger.

Keep in mind that the library defines a surface without a thickness. For more information, it is referred to the library. 

!!!note
    The implementation include currently Turbofan Kerosene only

## Pylon designer {#pylondesigner}
The pylon is the structural component to connect the engine to the aircraft. 

### Methods description

For the **pylon designer**, only one method is implemented:

 - *default* uses the `aircraftGeometry2` library 
 
In the current method, the mounting is attached to the beginning to the nacelle to the leading edge of the wing. The length is the engine length which is extruded to the wing. the profile is, likewise for the nacelle, defined in the _configXML_.

![Engine Mount](figures/engine_mount.svg)


!!!note 
    the implementation include currently Turbofan Kerosene only

## Mass analyzer {#massanalyzer}
Lastly, the mass properties for engine, nacelle and pylon are calculated separate for center of gravity, mass and inertia. 

### Methods description

Here, only one method is implemented:

 - *default* using: 
    - for engine
        - CG: calculating local CG assuming a circular cylinder
        - mass: empirical estimation
        - inertia: wrt. CG assuming solid cylinder
    - for nacelle & pylon
        - CG: calculating local CG with `aircraftGeometry2`lib
        - mass: empirical estimation
        - inertia: wrt. CG with `aircraftGeometry2`lib

!!!note 
    the implementation include currently Turbofan Kerosene only

